# -*- coding: utf-8 -*-

FAILURE_CLASSES = ("not classified", "fixed by commit")
INTERMITTENT_CLASSES = ("intermittent", "autoclassified intermittent")
TASK_FINAL_STATES = ("completed", "failed", "exception")
